module Parser(
    Parse
  , none
  , suceed
  , token
  , spot
  , (>*>)
  , use
  , use2
  , (*>)
  , (>*)
  , alt
  , list
  , some
  , parse
  ) where

type Parse alpha beta = [alpha]-> [(beta, [alpha])] 

-- Basisparser: erkennt nichts
none :: Parse alpha beta
none = const []

-- Basisparser: erkennt alles
suceed :: beta-> Parse alpha beta
suceed b inp = [(b, inp)]

-- Basisparser: erkennen einzelne Token
spot  :: (alpha-> Bool)-> Parse alpha alpha
spot p []     = []
spot p (x:xs) = if p x then [(x, xs)] else []

token :: Eq alpha=> alpha-> Parse alpha alpha
token t = spot (t ==)

-- Kombinator: Alternierung
infixl 3 `alt`
alt :: Parse alpha beta-> Parse alpha beta-> Parse alpha beta
alt p1 p2 i = p1 i ++ p2 i

-- Kombinator: Sequenzierung
infixl 5 >*>
(>*>) :: Parse alpha beta-> Parse alpha gamma-> Parse alpha (beta, gamma)
(>*>) p1 p2 i =
  concatMap (\(b, r)-> 
     map (\(c, s)-> ((b, c), s)) (p2 r)) (p1 i)

-- Kombinator: use (map fr Parser)
infix 4 `use`, `use2`
use :: Parse alpha beta-> (beta-> gamma)-> Parse alpha gamma
use p f i = map (\(o, r)-> (f o, r)) (p i)

use2 :: Parse alpha (beta, gamma)-> (beta-> gamma-> delta)-> Parse alpha delta
use2 p f = use p (uncurry f)

-- Sequenzierung rechts/links
infixl 5 *>, >*
(*>) :: Parse alpha beta-> Parse alpha gamma-> Parse alpha gamma
(>*) :: Parse alpha beta-> Parse alpha gamma-> Parse alpha beta
p1 *> p2 = p1 >*> p2 `use` snd
p1 >* p2 = p1 >*> p2 `use` fst

-- Kombinator: Kleene-Stern
list :: Parse alpha beta-> Parse alpha [beta]
list p = p >*> list p `use2` (:)
         `alt` suceed []

-- Kombinator: nicht-leere Liste
some :: Parse alpha beta-> Parse alpha [beta]
some p = p >*> list p `use2` (:)

-- Hauptfunktion:
parse :: Parse alpha beta-> [alpha]-> Either String beta 
parse p i = 
  case filter (null. snd) $ p i of
    []       -> Left "Input does not parse" 
    [(e, _)] -> Right e
    _        -> Left "Input is ambiguous"

-- Variante: erlaubt mehrdeute Parsierung, liefert leere Liste als Fehler
mparse :: Parse alpha beta-> [alpha]-> [beta]
mparse p i = 
  map fst $ filter (null. snd) $ p i 
